import logging
import io
import time
from tqdm import tqdm
import numpy as np

class TqdmToLogger(io.StringIO):
    """
        Output stream for TQDM which will output to logger module instead of
        the StdOut.
    """
    logger = None
    level = None
    buf = ''
    def __init__(self,logger,level=None):
        super(TqdmToLogger, self).__init__()
        self.logger = logger
        self.level = level or logging.INFO
    def write(self,buf):
        self.buf = buf.strip('\r\n\t ')
    def flush(self):
        self.logger.log(self.level, self.buf)
tqdm_logger = TqdmToLogger(
    logging.getLogger()
)

def get_stats(
    optimizer,
    num_iterations,
    optimizer_args,
    optimizer_additional_args={},
    transformer=None,
    verbose=True,
) -> dict:
    """
    ------------------------------------------------------
    Description:
        Return a dictionary with information about how the opmizer algorithm
        perform after N evaluations applied to opmizer function.

        The dictionary has the mean solution obtained, standard desviation, worst and
        best solution.

    Arguments:
        - optimizer: optimization class.
        - num_iterations: evaluation number, which is the number of times to
        applied class.optimize(args).
        - optimizer_args: Arguments necessary to perform.
        - optimizer_additional_args(optional): additional arguments passed to optimizer.
        - transformer: Function that return a float value. The fuction input is
            the best individual obtained after an execution.
        - verbose: display every solution and objective function.
    ------------------------------------------------------
    """
    data_by_execution = {"execution_time": [], "individual_x": [], "individual_f": []}
    for i in tqdm(range(num_iterations), file=tqdm_logger, miniters=4,):
        start_time = time.time()
        optimizer.optimize(*optimizer_args, **optimizer_additional_args)
        data_by_execution["execution_time"].append(time.time() - start_time)
        data_by_execution["individual_x"].append(optimizer.logger["best_individual"])
        function_value = optimizer.logger["best_f"]
        if transformer is not None:
            function_value = transformer(optimizer.logger["best_individual"])
        data_by_execution["individual_f"].append(function_value)

    ind_worst = np.argmax(data_by_execution["individual_f"])
    ind_best = np.argmin(data_by_execution["individual_f"])
    stats = {"Worst solution": {}, "Best solution": {}}

    stats["Worst solution"]["x"] = data_by_execution["individual_x"][ind_worst]
    stats["Best solution"]["x"] = data_by_execution["individual_x"][ind_best]

    stats["Worst solution"]["f"] = data_by_execution["individual_f"][ind_worst]
    stats["Best solution"]["f"] = data_by_execution["individual_f"][ind_best]
    stats["Mean"] = np.mean(data_by_execution["individual_f"])
    stats["Standard deviation"] = np.std(data_by_execution["individual_f"])
    stats["Median"] = np.median(data_by_execution["individual_f"])

    if verbose:
        stats.update(data_by_execution)

    return stats


class ContinuosFixer:
    """
    Description:
        When a solution is pass to this callback, the solution is setting as follow:
        Given an interval, values outside the interval are clipped to the interval edges.
        For example, if an interval of [0, 1] is specified, values smaller than 0 become 0,
        and values larger than 1 become 1.
    Arguments:
        - bounds: float list. The first element is the lower bound and
            the second element is the upper bound. Where could be arrays meaning
            the boundaries for every decision variable.
    """

    def __init__(self, bounds: list):
        self.bounds = bounds
        self.__doc__ = "continuos"

    def __call__(self, solution: np.ndarray, ind: int):
        return np.clip(solution[ind], self.bounds[0], self.bounds[1])


class NoneFixer:
    """
    Description:
        This callback doesn't anything. It only keeps as a standard of pyristic.
    Arguments:
        - None.
    """

    def __init__(self):
        self.__doc__ = "None"

    def __call__(self, solution: np.ndarray, ind: int):
        return solution



from copy import deepcopy
import numpy as np

__all__ = [
    "ProporcionalSampler",
    "RouletteSampler",
    "StochasticUniversalSampler",
    "DeterministicSampler",
    "TournamentSampler",
    "MergeSelector",
    "ReplacementSelector",
    "proporcional_sampling",
    "roulette_sampling",
    "stochastic_universal_sampling",
    "deterministic_sampling",
    "tournament_sampling",
    "get_expected_values",
    "get_candidates_by_aptitude",
]


def get_expected_values(aptitude: np.array) -> np.array:
    """
    Description:
        return the expected number of copies of every individual.
    """
    average_aptitude = np.sum(aptitude)
    number_individuals = len(aptitude)
    return aptitude / average_aptitude * number_individuals


"""
---------------------------------------------------------------------------------
                            |   Genetic Algorithms.   |
---------------------------------------------------------------------------------
"""


"""
---------------------------------------------------------------------------------
                                Parent selection.
---------------------------------------------------------------------------------
"""


def proporcional_sampling(expected_vals: np.ndarray) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        return an array with the indices from 0 up to n, where n
        is the number the elements of the input array.
    Arguments:
        probabilities: 1-D numpy array which is the probability of every
        individual based on the aptitude.
    About:
        If you want more information, check:

    ------------------------------------------------------
    """
    return np.arange(len(expected_vals))


def roulette_sampling(expected_vals: np.ndarray) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        selection algorithm O(nlogn) for genetic algorithm.
    Arguments:
        probabilities: 1-D numpy array which is the probability of every
        individual based on the aptitude.
    About:
        If you want more information, check:

    ------------------------------------------------------
    """
    expected_vals = get_expected_values(expected_vals)
    number_individuals = len(expected_vals)

    expected_cumulative_sum = np.cumsum(expected_vals)
    random_number = np.random.uniform(
        0.0, expected_cumulative_sum[-1], number_individuals
    )
    return np.searchsorted(expected_cumulative_sum, random_number)  # sample


def stochastic_universal_sampling(expected_vals: np.ndarray) -> np.ndarray:
    """
    ------------------------------------------------------
    Arguments:
        probabilities: 1-D numpy array which is the probability of every
        individual based on the aptitude.
    About:
        If you want more information, check:

    ------------------------------------------------------
    """
    expected_vals = get_expected_values(expected_vals)

    number_individuals = len(expected_vals)

    random_number = np.random.uniform(0, 1)
    current_sum = 0
    sample = []
    for i in range(number_individuals):
        current_sum += expected_vals[i]
        while random_number < current_sum:
            sample.append(i)
            random_number += 1
    return np.array(sample)


def deterministic_sampling(expected_vals: np.ndarray) -> np.ndarray:
    """
    ------------------------------------------------------
    Arguments:
        probabilities: 1-D numpy array which is the probability of every
        individual based on the aptitude.
    About:
        If you want more information, check:
    ------------------------------------------------------
    """
    expected_vals = get_expected_values(expected_vals)

    number_individuals = len(expected_vals)

    integer_part = np.array(expected_vals, dtype=int)
    indices = np.arange(number_individuals)
    sample = np.repeat(indices, integer_part)

    if len(sample) < number_individuals:
        float_part = expected_vals - integer_part
        ind = np.argpartition(float_part, len(sample) - number_individuals)
        ind = ind[len(sample) - number_individuals :]
        sample = np.concatenate((sample, indices[ind]))

    return sample[:number_individuals]


def tournament_sampling(
    expected_vals: np.ndarray, chunks: int = 2, prob: float = 1.0
) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        selection algorithm O(NP) for genetic algorithm where N is the population
        size and P is the number of chunks.
    Arguments:
        probabilities: 1-D numpy array which is the probability of every
        individual based on the aptitude.
        chunks: Integer value which is the number of chunks (by default is 2).
        prob: Float value which means if choose the max or min value in every chunk
        (by default is 1).
        About:
            If you want more information, check:

    ------------------------------------------------------
    """
    chunks = int(chunks)
    expected_vals = get_expected_values(expected_vals)
    num_items = len(expected_vals)
    indices = np.arange(num_items)
    chunks_ = np.ceil(num_items / chunks)

    sample = []
    ind = -1
    for tournament in range(chunks):
        groups = np.array_split(np.random.permutation(indices), chunks_)
        for group in groups:
            if prob >= np.random.rand():
                ind = np.argmax(expected_vals[group])
            else:
                ind = np.argmin(expected_vals[group])
            sample.append(group[ind])

    return np.array(sample)


"""
---------------------------------------------------------------------------------
                                Survivor selection.
---------------------------------------------------------------------------------
"""


def get_candidates_by_aptitude(value: np.ndarray, number_items: int) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Return m indices with the biggest fitness.
    Arguments:
        value: float array.
        number_items: int number of indices to return with the biggest value.
    ------------------------------------------------------
    """
    pseudo_sorted_array = np.argpartition(value, -1 * number_items)
    return pseudo_sorted_array[-1 * number_items :]


"""
---------------------------------------------------------------------------------
                                Classes
---------------------------------------------------------------------------------
"""

"""
----------------------------
    Parent Selection.
----------------------------
"""


class ProporcionalSampler:
    """
    Description:
      Selection class as method based on the proporcional_sampler.
    Arguments:
        - None
    """

    def __init__(self):
        self.__doc__ = "Proporcional sampling"

    def __call__(self, population_f: np.ndarray) -> np.ndarray:
        return proporcional_sampling(population_f)


class RouletteSampler:
    """
    Description:
      Selection class as method based on the roulette_sampler.
    Arguments:
        - None
    """

    def __init__(self):
        self.__doc__ = "Roulette sampling"

    def __call__(self, population_f: np.ndarray) -> np.ndarray:
        vals = np.copy(population_f)
        return roulette_sampling(vals)


class StochasticUniversalSampler:
    """
    Description:
      Selection class as method based on the stochastic_universal_sampler.
    Arguments:
        - None
    """

    def __init__(self):
        self.__doc__ = "Stochastic universal sampling"

    def __call__(self, population_f: np.ndarray) -> np.ndarray:
        vals = np.copy(population_f)
        return stochastic_universal_sampling(vals)


class DeterministicSampler:
    """
    Description:
      Selection class as method based on the deterministic_sampler.
    Arguments:
        - None
    """

    def __init__(self):
        self.__doc__ = "Deterministic sampling"

    def __call__(self, population_f: np.ndarray) -> np.ndarray:
        vals = np.copy(population_f)
        return deterministic_sampling(vals)


class TournamentSampler:
    """
    Description:
      Selection class as method based on the tournament_sampler.
    Arguments:
        - chunks: Integer value which is the number of chunks (by default is 2).
        - prob: Float value which means if choose the max or min value in every chunk
        (by default is 1).
    """

    def __init__(self, chunks_: int = 2, prob_: float = 1.0):
        self.chunks = chunks_
        self.prob = prob_
        self.__doc__ = f"Tournament sampling\n\t Arguments:\n\t\t-Chunks: {self.chunks}\n\t\t-prob: {self.prob}"

    def __call__(self, population_f: np.ndarray) -> np.ndarray:
        vals = np.copy(population_f)
        return tournament_sampling(vals, self.chunks, self.prob)


"""
----------------------------
    Survivor Selection.
----------------------------
"""


class MergeSelector:
    """
    Description:
      Selection survivors class as method based on the lambda + mu selection.
    Arguments:
        - None
    """

    def __init__(self):
        self.__doc__ = "Merge population"

    def __call__(
        self, parent_f: np.ndarray, offspring_f: np.ndarray, features: dict
    ) -> dict:
        """
        ------------------------------------------------------
        Description:
            survivor selection algorithm, where choose a individuals between parents and
            offspring that have great f for the next generation.
        Arguments:
            parent_f: 1-D numpy array.
            offspring_f: 1-D numpy array.
            features: dictionary where every key is a list of two numpy arrays, which means
                    key- [parent population feature , offspring population feature].
                Note: This features will be saved to the next generation.
            About:
                If you want more information, check:

        ------------------------------------------------------
        """
        result = {}

        tmp_f = np.concatenate((parent_f, offspring_f))
        indices = get_candidates_by_aptitude(tmp_f, len(parent_f))
        result["parent_population_f"] = tmp_f[indices]

        for feature in features.keys():
            if len(features[feature]) != 2:
                raise Exception(
                    (
                        "Lenght of list have to be 2 "
                        "(parent population feature and offspring population feature)."
                    )
                )
            tmp = np.concatenate((features[feature][0], features[feature][1]), axis=0)
            result[feature] = deepcopy(tmp[indices])
        return result


class ReplacementSelector:
    """
    Description:
      Selection survivors class as method based on the (lambda, mu) selection.
    Arguments:
        - None
    """

    def __init__(self):
        self.__doc__ = "Replacement population"

    def __call__(
        self, parent_f: np.ndarray, offspring_f: np.ndarray, features: dict
    ) -> dict:
        """
        ------------------------------------------------------
        Description:
            survivor selection algorithm, where choose m individuals in
            offspring that have great f for the next generation.
        Arguments:
            parent_f: 1-D numpy array.
            offspring_f: 1-D numpy array.
            features: dictionary where every key is a list of two numpy arrays, which means
                    key- [parent population feature , offspring population feature].
                Note: This features will be saved to the next generation.
            About:
                If you want more information, check:

        ------------------------------------------------------
        """
        assert len(parent_f) <= len(offspring_f)

        result = {}
        indices = get_candidates_by_aptitude(offspring_f, len(parent_f))
        result["parent_population_f"] = offspring_f[indices]

        for feature in features.keys():
            if len(features[feature]) != 2:
                raise Exception(
                    (
                        "Lenght of list have to be 2 "
                        "(parent population feature and offspring population feature)."
                    )
                )
            result[feature] = deepcopy(features[feature][1][indices])

        return result



import numpy as np

__all__ = [
    "InsertionMutator",
    "ExchangeMutator",
    "BoundaryMutator",
    "UniformMutator",
    "NoneUniformMutator",
    "BinaryMutator",
    "NoneMutator",
    "SigmaMutator",
    "MultSigmaAdaptiveMutator",
    "SingleSigmaAdaptiveMutator",
    "SigmaEpAdaptiveMutator",
    "sigma_ep_adaptive",
    "single_sigma_adaptive",
    "mult_sigma_adaptive",
    "mutation_by_sigma",
    "insertion_mutation",
    "insertion_mutation",
    "exchange_mutation",
    "boundary_mutation",
    "uniform_mutation",
    "none_uniform_mutation",
    "binary_mutation",
    "boundary_mutation_array",
    "uniform_mutation_array",
]


def sigma_ep_adaptive(population: np.ndarray, alpha: float) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Mutation operator for Evolutionary programming.
    ------------------------------------------------------
    """
    return population * (1.0 + alpha * np.random.normal(0, 1, size=population.shape))


def single_sigma_adaptive(population: np.ndarray, gamma: float) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Every individual has a sigma parameter which helps you to
        update in a fixed portion the individual.

        Note:
            This function is for Evolution Strategy search.
    Arguments:
        - X: Numpy array where every N-th row is an individual and
          M columns that are the decision variables.
        - gamma: Float value, additional parameter.

    ------------------------------------------------------
    """
    exponent = np.random.normal(0, 1) * gamma
    return np.exp(exponent) * population


def mult_sigma_adaptive(
    population: np.ndarray, gamma: float, gamma_prime: float
) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Every individual has a sigma for every decision variable which helps you to
        update in a fixed portion the individual.

        Note:
            This function is for Evolution Strategy search.
    Arguments:
        - X: Numpy array where every N-th row is an individual and
          M columns that are the decision variables.
        - gamma: Float value, additional parameter.
        - gamma_prime: Float value, additional parameter.
    ------------------------------------------------------
    """
    firts_ = np.random.normal(0, 1) * gamma_prime
    second_ = np.random.normal(0, 1, size=population.shape) * gamma

    exponent = firts_ + second_
    return population * np.exp(exponent)


def mutation_by_sigma(population: np.ndarray, sigma: np.ndarray) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Update function in the decision variables of every individual.

        Note:
            This function is for Evolution Strategy search.
    Arguments:
        - X: Numpy Matrix, where every N-th row is an individual and
          M columns that are the decision variables.
        - sigma: Numpy Array with N rows which means every row is a individual, M columns differ in
        what kind of sigma has chosen (single or multiple).
    ------------------------------------------------------
    """
    normal_value = np.random.normal(0, 1)
    return population + sigma * normal_value


def insertion_mutation(population: np.ndarray, n_elements: int = 1) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Discrete mutation operator. It takes a segment and the remaining elements
        are introduced.

    Arguments:
        - population: Numpy Matrix, where every N-th row is an individual and
          M columns that are the decision variables.
        - n_elements: Integer number (default 1). It indicates the number of
          elements in the segment.
    ------------------------------------------------------
    """
    n_elements = int(n_elements)
    num_individuals, decision_variables = population.shape
    x_mutated = np.ones((num_individuals, decision_variables))

    for ind in range(num_individuals):
        individual = np.full(decision_variables, np.inf)
        indices_elements = np.random.choice(
            decision_variables, n_elements, replace=False
        )
        elements = population[ind][indices_elements]
        indices_position = np.random.choice(
            decision_variables, n_elements, replace=False
        )
        individual[indices_position] = elements

        indices_elements.sort()
        remaining = []
        start_ = 0
        for decision_var in range(decision_variables):
            if (
                start_ < len(indices_elements)
                and decision_var == indices_elements[start_]
            ):
                start_ += 1
                continue
            remaining.append(population[ind][decision_var])

        individual[individual == np.inf] = remaining
        x_mutated[ind] = individual

    return x_mutated


def exchange_mutation(population: np.ndarray) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Discrete mutation operator. It takes two random positions
        and swap the values.

    Arguments:
        - population: Numpy Matrix, where every N-th row is an individual and
          M columns that are the decision variables.
    ------------------------------------------------------
    """
    num_individuals, decision_variables = population.shape
    tmp_x = np.copy(population)
    for ind in range(num_individuals):
        exchange_points = np.random.choice(decision_variables, 2, replace=False)
        first_decision_var = exchange_points[0]
        second_decision_var = exchange_points[1]

        tmp_point = tmp_x[ind][first_decision_var]
        tmp_x[ind][first_decision_var] = tmp_x[ind][second_decision_var]
        tmp_x[ind][second_decision_var] = tmp_point

    return tmp_x


def binary_mutation(population: np.ndarray, prob_mutation: float) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Binary mutation operator. This operator flip every position
        to 0 when the value is 1 or viceversa with a probability.

    Arguments:
        - population: Numpy Matrix, where every N-th row is an individual and
          M columns that are the decision variables.
        - prob_mutation: float number. It indicates the probability of flip.
    ------------------------------------------------------
    """
    assert 0 <= prob_mutation <= 1.0
    num_individuals, decision_variables = population.shape
    mutated_x = np.copy(population)
    for i_individual in range(num_individuals):
        for i_variable in range(decision_variables):
            if np.random.rand() < prob_mutation:
                num = 0
                if mutated_x[i_individual][i_variable] == 0:
                    num = 1
                mutated_x[i_individual][i_variable] = num

    return mutated_x


def boundary_mutation_array(
    population: np.ndarray, lower_bounds: list, upper_bounds: list
) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        continuous mutation operator. This operator select a
        random position of the individual and change the value to
        one of the boundaries.

        Note: this function is when every decision variable  has
        a different search space.
    Arguments:
        - population: Numpy Matrix, where every N-th row is an individual and
          M columns that are the decision variables.
        - lower_bounds: list of float numbers. The lower bound for the i-th
            decision variable.
        - upper_bounds: list of float numbers. The upper bound for the i-th
            decision variable.
    ------------------------------------------------------
    """
    num_individuals, decision_variables = population.shape
    mutated_x = np.copy(population)
    for ind in range(num_individuals):
        variable = np.random.randint(0, decision_variables)
        lower_bound = lower_bounds[ind]
        upper_bound = upper_bounds[ind]
        mutated_x[ind][variable] = lower_bound
        if np.random.rand() > 0.5:
            mutated_x[ind][variable] = upper_bound
    return mutated_x


def boundary_mutation(
    population: np.ndarray, lower_bound: float, upper_bound: float
) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        continuous mutation operator. This operator select a
        random position of the individual and change the value to
        one of the boundaries.

        Note: this function is when every decision variable  has
        the same search space.
    Arguments:
        - population: Numpy Matrix, where every N-th row is an individual and
          M columns that are the decision variables.
        - lower_bound: float number. The lower bound for the i-th
            decision variable.
        - upper_bound: float number. The upper bound for the i-th
            decision variable.
    ------------------------------------------------------
    """
    num_individuals, decision_variables = population.shape
    mutated_x = np.copy(population)
    for ind in range(num_individuals):
        variable = np.random.randint(0, decision_variables)
        mutated_x[ind][variable] = lower_bound
        if np.random.rand() > 0.5:
            mutated_x[ind][variable] = upper_bound
    return mutated_x


def uniform_mutation_array(
    population: np.ndarray, lower_bounds: list, upper_bounds: list
) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        continuous mutation operator. This operator select a
        random position of the individual and change the value by
        a random value between the boundaries for that i-th position.

        Note: this function is when every decision variable  has
        different search space.
    Arguments:
        - population: Numpy Matrix, where every N-th row is an individual and
          M columns that are the decision variables.
        - lower_bounds: array of float numbers. The lower bound for the i-th
            decision variable.
        - upper_bounds: array of float numbers. The upper bound for the i-th
            decision variable.
    ------------------------------------------------------
    """
    num_individuals, decision_variables = population.shape
    mutated_x = np.copy(population)
    for ind in range(num_individuals):
        variable = np.random.randint(0, decision_variables)
        lower_bound = lower_bounds[ind]
        upper_bound = upper_bounds[ind]
        mutated_x[ind][variable] = np.random.uniform(lower_bound, upper_bound)
    return mutated_x


def uniform_mutation(
    population: np.ndarray, lower_bound: float, upper_bound: float
) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        continuous mutation operator. This operator select a
        random position of the individual and change the value by
        a random value between the boundaries for that i-th position.

        Note: this function is when every decision variable  has
        the same search space.
    Arguments:
        - population: Numpy Matrix, where every N-th row is an individual and
          M columns that are the decision variables.
        - lower_bound: float number. The lower bound for the i-th
            decision variable.
        - upper_bound: float number. The upper bound for the i-th
            decision variable.
    ------------------------------------------------------
    """
    num_individuals, decision_variables = population.shape
    mutated_x = np.copy(population)
    for ind in range(num_individuals):
        variable = np.random.randint(0, decision_variables)
        mutated_x[ind][variable] = np.random.uniform(lower_bound, upper_bound)
    return mutated_x


def none_uniform_mutation(population: np.ndarray, sigma: float = 1.0) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        continuous mutation operator. This operator applies
        a random noise with a normal distribution between 0
        up to sigma for every decision variable.

    Arguments:
        - population: Numpy Matrix, where every N-th row is an individual and
          M columns that are the decision variables.
        - sigma: float number. it is the size of numbers acepted as
            noise.
    ------------------------------------------------------
    """
    num_individuals, decision_variables = population.shape
    mutated_x = np.copy(population)
    for ind in range(num_individuals):
        noise = np.random.normal(0, sigma, size=decision_variables)
        mutated_x[ind] *= noise

    return mutated_x


class InsertionMutator:
    """
    Description:
      Class mutation operator based on insertion_mutation.
    Arguments:
        - n_elements:
    """

    def __init__(self, n_elements: int = 1):
        self.n_elements = n_elements
        self.__doc__ = f"Insertion \n\t Arguments:\n\t\t -n_elements: {n_elements}"

    def __call__(self, population: np.ndarray) -> np.ndarray:
        return insertion_mutation(population, self.n_elements)


class ExchangeMutator:
    """
    Description:
      Class mutation operator based on exchange_mutation.
    Arguments:
        This method doesn't need arguments.
    """

    def __init__(self):
        self.__doc__ = "Exchange"

    def __call__(self, population: np.ndarray) -> np.ndarray:
        return exchange_mutation(population)


class BoundaryMutator:
    """
    Description:
      Class mutation operator based on boundary_mutation.
    Arguments:
        - bounds:
    """

    def __init__(self, bounds: list):
        self.bounds = bounds
        self.type_bound = list if isinstance(self.bounds[0], list) else float
        self.__doc__ = (
            "Boundary\n\t Arguments:\n\t\t"
            f" -Lower bound: {self.bounds[0]}\n\t\t -Upper bound: {self.bounds[1]}"
        )

    def __call__(self, population: np.ndarray) -> np.ndarray:
        if self.type_bound != list:
            return boundary_mutation(population, self.bounds[0], self.bounds[1])

        return boundary_mutation_array(population, self.bounds[0], self.bounds[1])


class UniformMutator:
    """
    Description:
      Class mutation operator based on unfirom_mutation.
    Arguments:
        - bounds:
    """

    def __init__(self, bounds: list):
        self.bounds = bounds
        self.type_bound = list if isinstance(self.bounds[0], list) else float
        self.__doc__ = (
            "Uniform\n\t Arguments:\n\t\t"
            f" -Lower bound: {self.bounds[0]}\n\t\t -Upper bound: {self.bounds[1]}"
        )

    def __call__(self, population: np.ndarray) -> np.ndarray:

        if self.type_bound != list:
            return uniform_mutation(population, self.bounds[0], self.bounds[1])

        return uniform_mutation_array(population, self.bounds[0], self.bounds[1])


class NoneUniformMutator:
    """
    Description:
      Class mutation operator based on none_uniform_mutation.
    Arguments:
        - sigma:
    """

    def __init__(self, sigma: float = 1.0):
        self.sigma = sigma
        self.__doc__ = f"Non Uniform\n\t Arguments:\n\t\t -Sigma: {self.sigma}"

    def __call__(self, population: np.ndarray) -> np.ndarray:
        return none_uniform_mutation(population, self.sigma)


class NoneMutator:
    """
    Description:
      Class mutation operator based on none_mutation.
    Arguments:
        This methos doesn't need arguments.
    """

    def __init__(self):
        self.__doc__ = "None"

    def __call__(self, population: np.ndarray) -> np.ndarray:
        return population


class BinaryMutator:
    """
    Description:
      Class mutation operator based on binary_mutation.
    Arguments:
        - prob_mutation: float number (default 0.2).
    """

    def __init__(self, prob_mutation: float = 0.2):
        self.__doc__ = (
            "Binary mutation \n\t Arguments:\n\t\t "
            f"- probability to flip: {prob_mutation}"
        )
        self.prob_mutation = prob_mutation

    def __call__(self, population: np.ndarray) -> np.ndarray:
        return binary_mutation(population, self.prob_mutation)


# Mutatio operator for EP.
class SigmaEpAdaptiveMutator:
    """
    Description:
      Class mutation operator based on sigma_ep_adaptive.
    Arguments:
        - decision_variables: integer number. It describe
            the number of decision variables.
        - alpha: float number. It means how much impact
            the noise applied to the individual.
    """

    def __init__(self, decision_variables: int, alpha: float):
        self._alpha = alpha
        self._length = int(decision_variables)
        self.__doc__ = f"Sigma EP.\n\t\t-Alpha: {self._alpha}"

    @property
    def length(self):
        """
        Return the number of decision variables.
        """
        return self._length

    def __call__(self, population: np.ndarray) -> np.ndarray:
        return sigma_ep_adaptive(population, self._alpha)


# Mutation operator for ES.
class SigmaMutator:
    """
    Description:
      Class mutation operator based on mutation_by_sigma.
    Arguments:
        - None
    """

    def __init__(self):
        self.__doc__ = "Sigma"

    def __call__(self, population: np.ndarray, sigma: np.ndarray) -> np.ndarray:
        return mutation_by_sigma(population, sigma)


class MultSigmaAdaptiveMutator:
    """
    Description:
      Class mutation operator based on mult_sigma_adaptive.
    Arguments:
        - None
    """

    def __init__(self, decision_variables: int):
        self.__doc__ = "Sigma mult"
        self._length = int(decision_variables)
        self._gamma_prime = 1 / np.sqrt(2 * self._length)
        self._gamma = 1 / np.sqrt(2 * np.sqrt(self._length))

    @property
    def length(self):
        """
        Return the number of decision variables.
        """
        return self._length

    def __call__(self, sigma: np.ndarray) -> np.ndarray:
        return mult_sigma_adaptive(sigma, self._gamma, self._gamma_prime)


class SingleSigmaAdaptiveMutator:
    """
    Description:
      Class mutation operator based on single_sigma_adaptive.
    Arguments:
        - None
    """

    def __init__(self):
        self.__doc__ = "Single Sigma"
        self._length = 1
        self._tau = 1 / np.sqrt(self._length)

    @property
    def length(self):
        """
        Return the number of decision variables.
        """
        return self._length

    def __call__(self, population: np.ndarray) -> np.ndarray:
        return single_sigma_adaptive(population, self._tau)



import typing
from random import random
import math
import numpy as np
import copy
from pyristic.utils.helpers import *

__all__ = ["SimulatedAnnealing"]


class SimulatedAnnealing:
    """
    ------------------------------------------------------
    Description:
        Optimization search algorithm inspired on simulated
        annealing algorithm.
    Arguments:
        - Function: The aptitude function to minimize.
        - Constraints: Array with boolean functions.
    ------------------------------------------------------
    """

    def __init__(
        self,
        function: typing.Callable[[np.ndarray], typing.Union[int, float]],
        constraints: list,
        neighbor_generator: typing.Callable[[np.ndarray], np.ndarray] = None,
    ):
        self.function = function
        self.constraints = constraints
        self.neighbor_generator = neighbor_generator
        # Search information.
        self.logger = {}
        self.logger["best_individual"] = None
        self.logger["best_f"] = None

    def __str__(self):
        printable = (
            "Simulated Annealing: \n "
            f"f(X) = {self.logger['best_f']} \n X = {self.logger['best_individual']} \n "
        )
        first = True

        for constraint in self.constraints:
            if constraint.__doc__ is not None:
                if first:
                    first = False
                    printable += "Constraints: \n "

                constraint(self.logger["best_individual"])
                printable += f"{constraint.__doc__} \n"
        return printable

    def optimize(
        self,
        initial_solution: typing.Union[np.ndarray, typing.Callable[[], np.ndarray]],
        initial_temperature: float,
        eps: float,
        **kwargs,
    ) -> None:
        """
        ------------------------------------------------------
        Description:
            Main function to find the best solution using Simulated Annealing strategy.
        Arguments:
            - initial_solution: Numpy array, represent the initial solution or function
                which generates a random initial solution (this solution should be a numpy array).
            - initial_temperature:  Floating value, it define how much allow worse solutions.
            - eps: it means what's the final temperature.
        ------------------------------------------------------
        """
        try:
            candidate = initial_solution()
        except TypeError:
            candidate = copy.deepcopy(initial_solution)

        self.logger["best_individual"] = copy.deepcopy(candidate)
        self.logger["best_f"] = self.function(copy.deepcopy(candidate))
        self.logger["temperature"] = initial_temperature

        f_candidate = self.function(candidate)
        while self.logger["temperature"] >= eps:
            neighbor = self.get_neighbor(candidate, **kwargs)

            if not self.__is_valid(neighbor):
                continue

            f_neighbor = self.function(neighbor)

            if f_neighbor < f_candidate or random() < self.acceptance_measure(
                f_neighbor, f_candidate, **kwargs
            ):
                candidate = neighbor
                f_candidate = f_neighbor

            self.logger["temperature"] = self.update_temperature(**kwargs)

            # Update best solution found.
            if f_candidate < self.logger["best_f"]:
                self.logger["best_f"] = f_candidate
                self.logger["best_individual"] = candidate

    def acceptance_measure(self, f_n: float, f_x: float, **kwargs) -> float:
        """
        Description:
            Metric used to evaluate how much confidence accept
            a solution.
        """
        temperature = self.logger["temperature"]
        return math.exp(-(f_n - f_x) / temperature)

    def __is_valid(self, solution: np.ndarray) -> bool:
        """
        ------------------------------------------------------
        Description:
            Check if the current solution is valid.
        ------------------------------------------------------
        """
        for constrain in self.constraints:
            if not constrain(solution):
                return False
        return True

    def update_temperature(self, **kwargs) -> float:
        """
        ------------------------------------------------------
        Description:

        ------------------------------------------------------
        """
        return self.logger["temperature"] * 0.99

    def get_neighbor(self, solution: np.ndarray, **kwargs) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
           Return only one solution which is a "random"
            variation of current solution.
        ------------------------------------------------------
        """
        if not self.neighbor_generator:
            raise NotImplementedError
        return self.neighbor_generator(solution)



import typing
import logging
from tqdm import tqdm
import numpy as np
from pyristic.utils.operators import selection, mutation, crossover, population_sample
from pyristic.utils.helpers import ContinuosFixer

__all__ = ["EvolutionStrategy"]
LOGGER = logging.getLogger()


class EvolutionStrategy:
    """
    ------------------------------------------------------
    Description:
        Initializing every variable necessary to the search.
    Arguments:
        - Function: Objective function to minimize.
        - Constraints: Constraints to be a feasible solution.
        - Bounds: bound for every variable, this should be a matrix 2 x N
        where N is the variables number. The first element is lowert limit
        and another one is the upper limit.
    ------------------------------------------------------
    """

    def __init__(
        self,
        function: typing.Callable[[np.ndarray], typing.Union[int, float]],
        decision_variables: int,
        constraints: list = [],
        bounds: list = [],
        config=None,
    ):
        self.aptitude_function = function
        self.constraints = constraints
        self.bounds = bounds
        self.decision_variables = decision_variables

        # Configuration.
        self.config_methods = {
            "init_population": population_sample.RandomUniformPopulation(
                self.decision_variables, self.bounds
            ),
            "mutation_operator": mutation.SigmaMutator(),
            "crossover_operator": crossover.DiscreteCrossover(),
            "survivor_selector": selection.MergeSelector(),
            "setter_invalid_solution": ContinuosFixer(self.bounds),
            "adaptive_crossover_operator": crossover.IntermediateCrossover(),
            "adaptive_mutation_operator": mutation.MultSigmaAdaptiveMutator(
                self.decision_variables
            ),
        }

        if config:
            self.config_methods.update(config.methods)

        # Search information.
        self.logger = {}
        self.logger["best_individual"] = None
        self.logger["best_f"] = None
        self.logger["current_iter"] = None
        self.logger["total_iter"] = None

    def __str__(self):
        printable = (
            "Evolution Strategy search: \n "
            f"F_a(X) = {self.logger['best_f']} \n X = {self.logger['best_individual']} \n "
        )
        first = True

        for constraint in self.constraints:
            if constraint.__doc__:

                if first:
                    first = False
                    printable += "Constraints: \n "

                constraint(self.logger["best_individual"])
                printable += f"{constraint.__doc__} \n"

        return printable

    def optimize(
        self,
        generations: int,
        population_size: int,
        offspring_size: int,
        eps_sigma: float = 0.001,
        verbose=True,
        **kwargs,
    ) -> None:
        """
        ------------------------------------------------------
        Description:
            The main function to find the best solution using tabu search.
        Arguments:
            - generations: integer that represent the total iterations.
            - population_size: the population that
        ------------------------------------------------------
        """
        generations = int(generations)
        population_size = int(population_size)
        offspring_size = int(offspring_size)
        # Reset global solution.
        self.logger["best_individual"] = None
        self.logger["best_f"] = None
        self.logger["current_iter"] = 0
        self.logger["total_iter"] = generations
        self.logger["parent_population_size"] = population_size
        self.logger["offspring_population_size"] = offspring_size
        self.logger["parent_population_x"] = self.initialize_population(**kwargs)
        self.logger["parent_population_sigma"] = self.initialize_step_weights(
            eps_sigma, **kwargs
        )
        self.logger["parent_population_f"] = np.apply_along_axis(
            self.aptitude_function, 1, self.logger["parent_population_x"]
        )

        try:
            for generation in tqdm(range(generations), disable=not verbose):

                # Crossover.
                first_parent_indices, second_parent_indices = self.__get_pairs(**kwargs)
                self.logger["offspring_population_x"] = self.crossover_operator(
                    first_parent_indices, second_parent_indices, **kwargs
                )

                self.logger["offspring_population_sigma"] = self.adaptive_crossover(
                    first_parent_indices, second_parent_indices, **kwargs
                )
                # mutate.
                self.logger["offspring_population_sigma"] = self.adaptive_mutation(
                    **kwargs
                )
                self.logger["offspring_population_x"] = self.mutation_operator(**kwargs)

                self.__set_invalid_individuals()
                next_generation = self.survivor_selection(**kwargs)
                self.logger["parent_population_x"] = next_generation[
                    "parent_population_x"
                ]
                self.logger["parent_population_sigma"] = next_generation[
                    "parent_population_sigma"
                ]
                self.logger["parent_population_f"] = next_generation[
                    "parent_population_f"
                ]
                self.logger["current_iter"] += 1

        except KeyboardInterrupt:
            LOGGER.error("Interrupted, saving best solution found so far.")

        ind = np.argmin(self.logger["parent_population_f"])

        self.logger["best_individual"] = self.logger["parent_population_x"][ind]
        self.logger["best_f"] = self.logger["parent_population_f"][ind]

    def initialize_step_weights(self, eps_sigma: float, **kwargs) -> np.ndarray:
        """
        Description:
            Initialize the size of the steps for every individual.
        """
        steps = np.random.uniform(
            0,
            1,
            size=(
                self.logger["parent_population_size"],
                self.config_methods["adaptive_mutation_operator"].length,
            ),
        )
        return np.maximum(steps, eps_sigma)

    def initialize_population(self, **kwargs) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
            How to distribute the population. This function create a
            population using uniform distribution.
            This should return an matrix of size (M x N)
            where M is the number of population and N is the number of
            variables.
        Arguments:
            -size_: Tnteger n where n is the number of variables about the problem.
        ------------------------------------------------------
        """
        return self.config_methods["init_population"](
            self.logger["parent_population_size"]
        )

    def fixer(self, ind: int) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
        Function which helps to move solution in a valid region.
        Arguments:
            - ind: index of individual.
        ------------------------------------------------------
        """
        return self.config_methods["setter_invalid_solution"](
            self.logger["offspring_population_x"], ind
        )

    def crossover_operator(
        self, parent_ind1: np.ndarray, parent_ind2: np.ndarray, **kwargs
    ) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
            Apply the crossover operator selected by the configuration.
        ------------------------------------------------------
        """
        return self.config_methods["crossover_operator"](
            self.logger["parent_population_x"], parent_ind1, parent_ind2
        )

    def adaptive_crossover(
        self, parent_ind1: np.ndarray, parent_ind2: np.ndarray, **kwargs
    ) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
            Apply the adaptive crossover operator selected by the configuration.
        ------------------------------------------------------
        """
        return self.config_methods["adaptive_crossover_operator"](
            self.logger["parent_population_sigma"], parent_ind1, parent_ind2
        )

    def mutation_operator(self, **kwargs) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
            The current population is updated by specific change
            using the adaptive control.
            This function should mutate the population.
        ------------------------------------------------------
        """
        return self.config_methods["mutation_operator"](
            self.logger["offspring_population_x"],
            self.logger["offspring_population_sigma"],
        )

    def adaptive_mutation(self, **kwargs) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
            Apply the adaptive mutation operator selected by the configuration.
        ------------------------------------------------------
        """
        return self.config_methods["adaptive_mutation_operator"](
            self.logger["offspring_population_sigma"]
        )

    def survivor_selection(self, **kwargs) -> dict:
        """
        ------------------------------------------------------
        Description:
            Apply the survivor selection method selected by the configuration.
        ------------------------------------------------------
        """
        individuals = {}
        individuals["parent_population_x"] = [
            self.logger["parent_population_x"],
            self.logger["offspring_population_x"],
        ]
        individuals["parent_population_sigma"] = [
            self.logger["parent_population_sigma"],
            self.logger["offspring_population_sigma"],
        ]

        return self.config_methods["survivor_selector"](
            self.logger["parent_population_f"],
            self.logger["offspring_population_f"],
            individuals,
        )

    # -----------------------------------------------------
    # Private functions.
    # -----------------------------------------------------
    def __is_invalid(self, individual: np.ndarray) -> bool:
        """
        ------------------------------------------------------
        Description:
            Check if the current solution is invalid.
        ------------------------------------------------------
        """
        for constraint in self.constraints:
            if not constraint(individual):
                return True
        return False

    def __get_pairs(self, **kwargs):
        parent_ind1 = np.random.randint(
            self.logger["parent_population_size"],
            size=(self.logger["offspring_population_size"],),
        )
        parent_ind2 = np.random.randint(
            self.logger["parent_population_size"],
            size=(self.logger["offspring_population_size"],),
        )

        return parent_ind1, parent_ind2

    def __set_invalid_individuals(self):
        # Fixing solutions and getting aptitude.
        f_offspring = []
        for i in range(len(self.logger["offspring_population_x"])):
            if self.__is_invalid(self.logger["offspring_population_x"][i]):
                self.logger["offspring_population_x"][i] = self.fixer(i)
            f_offspring.append(
                self.aptitude_function(self.logger["offspring_population_x"][i])
            )

        self.logger["offspring_population_f"] = np.array(f_offspring)



from .GeneticAlgorithm_search import Genetic
from .EvolutionStrategy_search import EvolutionStrategy
from .EvolutiveProgramming_search import EvolutionaryProgramming
from .SimulatedAnnealing_search import SimulatedAnnealing
from .Tabu_search import TabuSearch



import typing
import numpy as np

__all__ = [
    "OptimizerConfig",
    "GeneticConfig",
    "EvolutionStrategyConfig",
    "EvolutionaryProgrammingConfig",
]


class OptimizerConfig:
    """
    Description:
        General configuration for the evolutionary
        optimization algorithms of pyristic.
    Arguments:
        - None
    """

    def __init__(self):
        self.methods = {}

    def cross(
        self,
        crossover_: typing.Callable[[np.ndarray, np.ndarray, np.ndarray], np.ndarray],
    ):
        """
        Description:
            This method update the current crossover operator
            which the algorithm apply. By default the configuration hasn't
            a crossover operator.
        """
        self.methods["crossover_operator"] = crossover_
        return self

    def mutate(self, mutate_: typing.Callable[[np.ndarray], np.ndarray]):
        """
        Description:
            This method update the current mutator operator
            which the algorithm apply. By default the configuration hasn't
            a mutation operator.
        """
        self.methods["mutation_operator"] = mutate_
        return self

    def survivor_selection(
        self, survivor_function: typing.Callable[[np.ndarray, np.ndarray, dict], dict]
    ):
        """
        Description:
            This method update the current suvivor selection method
            which the algorithm apply. By default the configuration hasn't
            a survivor method.
        """
        self.methods["survivor_selector"] = survivor_function
        return self

    def fixer_invalide_solutions(
        self, fixer_function: typing.Callable[[np.ndarray], np.ndarray]
    ):
        """
        Description:
            This method update the current method to fix the solution
            if this solution doesn't accomplish the constraints
            which the algorithm apply. By default the configuration hasn't
            a fix method.
        """
        self.methods["setter_invalid_solution"] = fixer_function
        return self

    def initialize_population(self, init_function: typing.Callable):
        """
        Description:
            This function update the current method to initialize the population
            of our algorithm. By default the configuration hasn't a method.
        """
        self.methods["init_population"] = init_function
        return self

    def __str__(self):
        """
        Description:
            attach the printable information of crossover, mutation, survivor selection.
        """
        printable = "--------------------------------\n\tConfiguration\n--------------------------------\n"
        for key, func in self.methods.items():
            printable += f"{key} - {func.__doc__}\n"
        return printable


class GeneticConfig(OptimizerConfig):
    """
    Description:
        Auxiliar class that keeps the methods need to execute the
        genetic algorithm search. This configuration class store the following methods:
        fixer solutions, parent selection, survivor strategy, crossover operator and
        mutation operator.
    Arguments:
        - None.
    """

    def __init__(self):
        super().__init__()

    def parent_selection(self, parent_function: typing.Callable):
        """
        Description:
            It saves in memory the parent selection method.
        Arguments:
            - parent_function: the callback need to perform the parent selection step.
        """
        self.methods["parent_selector"] = parent_function
        return self


class EvolutionStrategyConfig(OptimizerConfig):
    """
    Description:
        Auxiliar class that keeps the methods need to execute the
        evolutionary strategy algorithm search. This configuration class store the
        following methods: fixer solutions, survivor strategy, crossover operator,
        mutation operator, adaptive mutation and adaptive crossover operator.
    Arguments:
        - None.
    """

    def __init__(self):
        super().__init__()

    def adaptive_crossover(self, adaptive_crossover_function: typing.Callable):
        """
        Description:
            It saves in memory the adaptive crossover method.
        Arguments:
            - adaptive_crossover_function: the callback need to perform
            the adaptation crossover step.
        """
        self.methods["adaptive_crossover_operator"] = adaptive_crossover_function
        return self

    def adaptive_mutation(self, adaptive_mutation_function: typing.Callable):
        """
        Description:
            It saves in memory the adaptive mutation method.
        Arguments:
            - adaptive_mutation_function: the callback need to perform
            the adaptive mutation step.
        """
        self.methods["adaptive_mutation_operator"] = adaptive_mutation_function
        return self


class EvolutionaryProgrammingConfig(OptimizerConfig):
    """
    Description:
        Auxiliar class that keeps the methods need to execute the
        evolutionary strategy algorithm search. This configuration class store the
        following methods: fixer solutions, survivor strategy, mutation operator,
        adaptive mutation.
    Arguments:
        - None.
    """

    def __init__(self):
        super().__init__()

    def adaptive_mutation(self, adaptive_mutation_function: typing.Callable):
        """
        Description:
            It saves in memory the adaptive mutation method.
        Arguments:
            - adaptive_mutation_function: the callback need to perform
            the adaptive mutation step.
        """
        self.methods["adaptive_mutation_operator"] = adaptive_mutation_function
        return self



import typing
import logging
from tqdm import tqdm
import numpy as np
from pyristic.utils.helpers import ContinuosFixer
from pyristic.utils.operators import mutation, selection, population_sample

__all__ = ["EvolutionaryProgramming"]
LOGGER = logging.getLogger()


class EvolutionaryProgramming:
    """
    ------------------------------------------------------
    Description:
        Initializing every variable necessary to the search.
    Arguments:
        - Function: Objective function to minimize.
        - Decision variables: Number of decision variables.
        - Constraints: Constraints to be a feasible solution.
        - Bounds: Bounds for each variable. This should be a matrix 2 x N
        where N is the variables number. The first element is the lower limit,
        and another one is the upper limit.
        - Config: EvolutationaryProgrammingConfig, which changes the behavior of
        some operators.
    ------------------------------------------------------
    """

    def __init__(
        self,
        function: typing.Callable[[np.ndarray], typing.Union[int, float]],
        decision_variables: int,
        constraints: list = [],
        bounds: list = [],
        config=None,
    ):
        self.aptitude_function = function
        self.constraints = constraints
        self.bounds = bounds
        self.decision_variables = decision_variables

        # Setting operators.
        self.config_methods = {
            "init_population": population_sample.RandomUniformPopulation(
                self.decision_variables, self.bounds
            ),
            "mutation_operator": mutation.SigmaMutator(),
            "survivor_selector": selection.MergeSelector(),
            "setter_invalid_solution": ContinuosFixer(self.bounds),
            "adaptive_mutation_operator": mutation.SigmaEpAdaptiveMutator(
                self.decision_variables, 0.5
            ),
        }
        if config:
            self.config_methods.update(config.methods)

        # Global information.
        self.logger = {}
        self.logger["best_individual"] = None
        self.logger["best_f"] = None
        self.logger["current_iter"] = None
        self.logger["total_iter"] = None
        self.logger["parent_population_size"] = None
        self.logger["parent_population_x"] = None
        self.logger["parent_population_sigma"] = None
        self.logger["parent_population_f"] = None
        self.logger["offspring_population_x"] = None
        self.logger["offspring_population_sigma"] = None
        self.logger["offspring_population_f"] = None

    def __str__(self):
        printable = (
            "Evolutive Programming search: \n "
            f"F_a(X) = {self.logger['best_f']} \n X = {self.logger['best_individual']} \n "
        )
        first = True

        for constraint in self.constraints:
            if constraint.__doc__ is not None:
                if first:
                    first = False
                    printable += "Constraints: \n "

                constraint(self.logger["best_individual"])
                printable += f"{constraint.__doc__} \n"

        return printable

    def optimize(
        self, generations: int, size_population: int, verbose=True, **kwargs
    ) -> None:
        """
        ------------------------------------------------------
        Description:
            The main function to find the best solution using Evolutionary programming.
        Arguments:
            - Generations: Number of iterations.
            - Population size: Number of individuals.
        ------------------------------------------------------
        """
        generations = int(generations)
        size_population = int(size_population)
        # Reset global solution.
        self.logger["current_iter"] = 0
        self.logger["total_iter"] = generations
        self.logger["parent_population_size"] = size_population
        self.logger["best_individual"] = None
        self.logger["best_f"] = None

        # Initial population.

        self.logger["parent_population_x"] = self.initialize_population(**kwargs)
        self.logger["parent_population_sigma"] = self.initialize_step_weights(**kwargs)
        self.logger["parent_population_f"] = np.apply_along_axis(
            self.aptitude_function, 1, self.logger["parent_population_x"]
        )

        try:
            for generation in tqdm(range(generations), disable=not verbose):
                # Mutation.
                self.logger["offspring_population_sigma"] = self.adaptive_mutation(
                    **kwargs
                )

                self.logger["offspring_population_x"] = self.mutation_operator(**kwargs)

                # Fixing solutions and getting aptitude.
                f_offspring = []
                for ind in range(len(self.logger["offspring_population_x"])):
                    if self.__is_invalid(self.logger["offspring_population_x"][ind]):
                        self.logger["offspring_population_x"][ind] = self.fixer(ind)
                    f_offspring.append(
                        self.aptitude_function(
                            self.logger["offspring_population_x"][ind]
                        )
                    )
                self.logger["offspring_population_f"] = np.array(f_offspring)

                # Survivor selection.
                next_generation = self.survivor_selection(**kwargs)
                self.logger["parent_population_x"] = next_generation[
                    "parent_population_x"
                ]
                self.logger["parent_population_sigma"] = next_generation[
                    "parent_population_sigma"
                ]
                self.logger["parent_population_f"] = next_generation[
                    "parent_population_f"
                ]

                self.logger["current_iter"] += 1

        except KeyboardInterrupt:
            LOGGER.error("Interrupted, saving best solution found so far.")

        ind = np.argmin(self.logger["parent_population_f"])
        self.logger["best_individual"] = self.logger["parent_population_x"][ind]
        self.logger["best_f"] = self.logger["parent_population_f"][ind]

    def mutation_operator(self, **kwargs) -> None:
        """
        ------------------------------------------------------
        Description:
            A new population is created from the current population.
            This function applies small changes to the decision variables
            and the step sizes.
        ------------------------------------------------------
        """
        return self.config_methods["mutation_operator"](
            self.logger["parent_population_x"], self.logger["parent_population_sigma"]
        )

    def adaptive_mutation(self, **kwargs) -> None:
        """
        ------------------------------------------------------
        Description:
            Apply the adaptive mutation operator selected by the configuration.
        ------------------------------------------------------
        """
        return self.config_methods["adaptive_mutation_operator"](
            self.logger["parent_population_sigma"]
        )

    def survivor_selection(self, **kwargs) -> dict:
        """
        ------------------------------------------------------
        Description:
            Apply the survivor selection method selected by the configuration.
        ------------------------------------------------------
        """
        individuals = {}
        individuals["parent_population_x"] = [
            self.logger["parent_population_x"],
            self.logger["offspring_population_x"],
        ]
        individuals["parent_population_sigma"] = [
            self.logger["parent_population_sigma"],
            self.logger["offspring_population_sigma"],
        ]

        return self.config_methods["survivor_selector"](
            self.logger["parent_population_f"],
            self.logger["offspring_population_f"],
            individuals,
        )

    def initialize_step_weights(self, **kwargs) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
            Create the steps for every individual.
        ------------------------------------------------------
        """
        steps = np.random.uniform(
            0, 1, size=(self.logger["parent_population_size"], self.decision_variables)
        )
        return np.maximum(steps, 0.00001)

    def initialize_population(self, **kwargs) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
            This function creates a population using a uniform distribution.
            It returns a matrix of size (n x m), where n is the size population
            and m is the number of variables.
        Arguments:
            -n: The population size.
        ------------------------------------------------------
        """
        return self.config_methods["init_population"](
            self.logger["parent_population_size"]
        )

    def fixer(self, ind: int) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
        This function helps to move one solution to the feasible region.
        Arguments:
            - ind: Index of the individual.
        ------------------------------------------------------
        """
        return self.config_methods["setter_invalid_solution"](
            self.logger["offspring_population_x"], ind
        )

    def __is_invalid(self, individual: np.ndarray) -> bool:
        """
        ------------------------------------------------------
        Description:
            This function checks if the current solution is unfeasible
        ------------------------------------------------------
        """
        for constraint in self.constraints:
            if not constraint(individual):
                return True
        return False



import typing
import logging
import numpy as np

__all__ = ["RandomUniformPopulation", "RandomPermutationPopulation"]

LOGGER = logging.getLogger()


class RandomUniformPopulation:
    """
    Description:
      Creates a matrix where every row represents an individual.
      Those individuals are created using an uniform distribution.
    Arguments:
        - num_variables: Integer that represents the number of variables
                         of the problem.
        - bounds: Uni dimensional array with two integer. The first integer
                  is the lower bound and the second one is the upper bound or
                  Array with two array, every array represent the bounds by variable.
    """

    def __init__(
        self, num_variables: int, bounds: typing.Union[list[float], list[list[int]]]
    ) -> None:
        self.num_variables = num_variables
        if isinstance(num_variables, float):
            LOGGER.warning(
                "WARNING: The num_variable argument is floating number, updating to integer."
            )
            self.num_variables = int(num_variables)
        self.bounds = bounds
        self.__doc__ = f"""
            Generate random numbers using uniform distribution.
            Arguments:
                bounds: {self.bounds}
                num_variables: {self.num_variables}
        """

    def __call__(self, num_individulas: int) -> np.ndarray:
        return np.random.uniform(
            self.bounds[0], self.bounds[1], size=(num_individulas, self.num_variables)
        )


class RandomPermutationPopulation:
    """
    Description:
      Creates a matrix where every row represents an individual.
      Those individuals are created doing a new permutation of the
      first [0,...,n] elements.
    Arguments:
        - x: int or array_like. If x is an integer, randomly permute np.arange(x).
             If x is an array, make a copy and shuffle the elements randomly.
        - bounds: Uni dimensional array with two integer. The first integer
                  is the lower bound and the second one is the upper bound or
                  Array with two array, every array represent the bounds by variable.
    """

    def __init__(self, x: int) -> None:
        self.x = x
        if isinstance(x, float):
            LOGGER.warning(
                "WARNING: The x argument is floating number, updating to integer."
            )
            self.x = int(x)
        self.__doc__ = f"""
            Generate random permutations.
            Arguments: 
                x: {self.x}
        """

    def __call__(self, num_individuals: int) -> np.ndarray:
        individuals = []

        for _ in range(num_individuals):
            individuals += [np.random.permutation(self.x)]

        return np.array(individuals)



import numpy as np
import math


"""
--------------------------------------------------------------------------
                            Beale function.
--------------------------------------------------------------------------
"""


def beale_function(X: np.ndarray) -> float:
    a = (1.5 - X[0] + X[0] * X[1]) ** 2
    b = (2.25 - X[0] + X[0] * X[1] ** 2) ** 2
    c = (2.625 - X[0] + X[0] * X[1] ** 3) ** 2
    return a + b + c


def constraint1_beale(X: np.ndarray) -> bool:
    for i in range(len(X)):
        if -4.5 > X[i] or X[i] > 4.5:
            return False
    # If you want to see the result.
    constraint1_beale.__doc__ = (
        "x1: -4.5 <= {:.2f} <= 4.5 \n x2: -4.5 <= {:.2f} <= 4.5".format(X[0], X[1])
    )
    return True


beale_constraints = [constraint1_beale]
beale_bounds = [-4.5, 4.5]

beale_ = {
    "function": beale_function,
    "constraints": beale_constraints,
    "bounds": beale_bounds,
    "decision_variables": 2,
}

"""
--------------------------------------------------------------------------
                            Ackley function.
--------------------------------------------------------------------------
"""


def ackley_function(X: np.ndarray) -> float:
    n = len(X)
    square_sum = (1 / n) * np.sum(X * X)
    trigonometric_sum = (1 / n) * np.sum(np.cos(2 * np.pi * X))

    return (
        -20 * np.exp(-0.2 * np.sqrt(square_sum)) - np.exp(trigonometric_sum) + 20 + np.e
    )


def constraint1_ackley(X: np.ndarray) -> bool:
    str_ = ""
    valid = True
    for i in range(len(X)):
        if -30 > X[i] or X[i] > 30:
            valid = False
        str_ += "x{}: -30 <= {:.2f} <= 30 \n ".format(i + 1, X[i])

    # Important if you want to see the result.
    constraint1_ackley.__doc__ = str_
    return valid


ackley_constraints = [constraint1_ackley]
ackley_bounds = [-30.0, 30.0]

ackley_ = {
    "function": ackley_function,
    "constraints": ackley_constraints,
    "bounds": ackley_bounds,
    "decision_variables": 10,
}

"""
--------------------------------------------------------------------------
                            Bukin N.6 function.
--------------------------------------------------------------------------
"""


def bukin_function(X: np.ndarray) -> float:
    A = np.sqrt(np.abs(X[1] - 0.01 * X[0] ** 2))
    B = np.abs(X[0] + 10)
    return 100.0 * A + 0.01 * B


def constraint1_bukin(solution: np.ndarray) -> bool:
    str_ = ""
    str_ += f"x{1}: -15 <= {solution[0]:.2f} <= -5 \n "
    str_ += f"x{2}: -3 <= {solution[1]:.2f} <= 3 \n "
    constraint1_bukin.__doc__ = str_
    if -15 > solution[0] or solution[0] > -5:
        return False
    if -3 > solution[1] or solution[1] > 3:
        return False
    return True


bukin_constraints = [constraint1_bukin]
bukin_bounds = [[-15, -3], [-5, 3]]  # [[Lower bounds] , [Upper bounds]]

bukin_ = {
    "function": bukin_function,
    "constraints": bukin_constraints,
    "bounds": bukin_bounds,
    "decision_variables": 2,
}

"""
--------------------------------------------------------------------------
                            Himmelblau's function.
--------------------------------------------------------------------------
"""


def himmelblau_function(X: np.ndarray) -> float:
    A = (X[0] ** 2 + X[1] - 11) ** 2
    B = (X[0] + X[1] ** 2 - 7) ** 2
    return A + B


def constraint1_Himmelblau(solution: np.ndarray) -> bool:
    str_ = ""
    valid = True
    for ind, decision_var in enumerate(solution):
        if -5 > decision_var or decision_var > 5:
            valid = False
        str_ += f"x{ind+1}: -5 <= {decision_var:.2f} <= 5 \n "

    # Important if you want to see the result.
    constraint1_Himmelblau.__doc__ = str_
    return valid


Himmelblau_constraints = [constraint1_Himmelblau]
Himmelblau_bounds = [-5, 5]

Himmelblau_ = {
    "function": himmelblau_function,
    "constraints": Himmelblau_constraints,
    "bounds": Himmelblau_bounds,
    "decision_variables": 2,
}



from .helpers import get_stats
from .evolutionary_config import (
    GeneticConfig,
    EvolutionaryProgrammingConfig,
    EvolutionStrategyConfig,
)
from .test_function import beale_, ackley_, bukin_, Himmelblau_



import typing
import logging
from tqdm import tqdm
import numpy as np
from pyristic.utils.operators import population_sample

__all__ = ["Genetic"]
LOGGER = logging.getLogger()


class Genetic:
    """
    ------------------------------------------------------
    Description:
        Optimization search algorithm inspired on genetic
        algorithms from evolutionary algorithms field.
    Arguments:
        - Function: The aptitude function to minimize.
        - Constraints: Array with boolean functions.
        - Bounds: bound for every variable, this should be a matrix 2 x N
            where N is the variables number. The first element is lowert limit
            and another one is the upper limit.
    ------------------------------------------------------
    """

    def __init__(
        self,
        function: typing.Callable[[np.ndarray], typing.Union[int, float]],
        decision_variables: int,
        constraints: list = [],
        bounds: list = [],
        config=None,
    ):
        # Information about problem.
        self.aptitude_function = function
        self.constraints = constraints
        self.bounds = bounds
        self.decision_variables = decision_variables  # Decision variables.

        # Operators.
        self.config_methods = {
            "init_population": population_sample.RandomUniformPopulation(
                self.decision_variables, self.bounds
            ),
        }
        if config:
            self.config_methods.update(config.methods)
        # Search information.
        self.logger = {}
        self.logger["best_individual"] = None
        self.logger["best_f"] = None
        self.logger["current_iter"] = None
        self.logger["total_iter"] = None
        self.logger["parent_population_size"] = None

    def __str__(self):
        printable = (
            "Genetic search: \n F_a(X) = "
            f"{self.logger['best_f']} \n X = {self.logger['best_individual']} \n "
        )
        first = True

        for constraint in self.constraints:
            if constraint.__doc__ is not None:

                if first:
                    first = False
                    printable += "Constraints: \n "

                constraint(self.logger["best_individual"])
                printable += f"{constraint.__doc__} \n"

        return printable

    # -----------------------------------------------------
    # Public functions.
    # -----------------------------------------------------
    def optimize(
        self,
        generations: int,
        size_population: int,
        cross_percentage: float = 1.0,
        mutation_percentage: float = 1.0,
        verbose: bool = True,
        **kwargs,
    ) -> None:
        """
        ------------------------------------------------------
        Description:
            The main function to find the best solution using tabu search.
        Arguments:
            -generations: number of iterations.
            -population: Number of new solutions.
        ------------------------------------------------------
        """
        assert 0 <= cross_percentage <= 1.0
        assert 0 <= mutation_percentage <= 1.0
        generations = int(generations)
        size_population = int(size_population)
        # Reset global information.
        self.logger["current_iter"] = 0
        self.logger["total_iter"] = generations
        self.logger["parent_population_size"] = size_population
        self.logger["cross_percentage"] = cross_percentage
        self.logger["mutation_percentage"] = mutation_percentage
        self.logger["best_f"] = None
        self.logger["best_individual"] = None

        # Initial population.
        self.logger["parent_population_x"] = self.initialize_population(**kwargs)
        self.logger["parent_population_f"] = np.apply_along_axis(
            self.aptitude_function, 1, self.logger["parent_population_x"]
        )

        try:
            for g in tqdm(range(generations), disable=not verbose):
                # Parent selection.
                parent_ind = self.parent_selection(**kwargs)
                first_parent_indices, second_parent_indices = self.__get_pairs(
                    parent_ind
                )
                # Cross.
                self.__cross_individuals(first_parent_indices, second_parent_indices)
                # mutate.
                self.__mutate_individuals()

                self.__set_invalid_individuals()

                # Survivor selection.
                next_generation = self.survivor_selection(**kwargs)
                self.logger["parent_population_x"] = next_generation["population"]
                self.logger["parent_population_f"] = next_generation[
                    "parent_population_f"
                ]

                self.logger["current_iter"] += 1

        except KeyboardInterrupt:
            LOGGER.error("Interrupted, saving best solution found so far.")

        ind = np.argmin(self.logger["parent_population_f"])
        self.logger["best_individual"] = self.logger["parent_population_x"][ind]
        self.logger["best_f"] = self.logger["parent_population_f"][ind]

    def initialize_population(self, **kwargs) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
            How to distribute the population. This function create a
            population using uniform distribution.
            This should return an matrix of size (M x N)
            where M is the number of population and N is the number of
            variables.
        Arguments:
            -size_: tuple with two integers (m,n)
            where m is the number generated of new solutions and
            n is the number of variables about the problem.
        ------------------------------------------------------
        """
        return self.config_methods["init_population"](
            self.logger["parent_population_size"]
        )

    def fixer(self, ind: int) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
        Function which helps to move solution in a valid region.
        Arguments:
            -X: numpy array, this one is the current population.
            -ind: index of individual.
        ------------------------------------------------------
        """
        if "setter_invalid_solution" not in self.config_methods:
            raise NotImplementedError
        return self.config_methods["setter_invalid_solution"](
            self.logger["offspring_population_x"], ind
        )

    def mutation_operator(self, indices, **kwargs):
        """
        ------------------------------------------------------
        Description:
            Apply the mutation operator selected by the configuration.
        Arguments:
            - Indices: the indices (or boolean array) of individuals in offspring_population_x
            to mutate.
        ------------------------------------------------------
        """
        if "mutation_operator" not in self.config_methods:
            raise NotImplementedError
        return self.config_methods["mutation_operator"](
            self.logger["offspring_population_x"][indices]
        )

    def crossover_operator(
        self, parent_ind1: np.ndarray, parent_ind2: np.ndarray, **kwargs
    ):
        """
        ------------------------------------------------------
        Description:
            Apply the crossover operator selected by the configuration.
        ------------------------------------------------------
        """
        if "crossover_operator" not in self.config_methods:
            raise NotImplementedError
        return self.config_methods["crossover_operator"](
            self.logger["parent_population_x"], parent_ind1, parent_ind2
        )

    def survivor_selection(self, **kwargs) -> dict:
        """
        ------------------------------------------------------
        Description:
            Apply the selection survivors method by the configuration.
        ------------------------------------------------------
        """
        if "survivor_selector" not in self.config_methods:
            raise NotImplementedError

        individuals = {}
        individuals["population"] = [
            self.logger["parent_population_x"],
            self.logger["offspring_population_x"],
        ]

        return self.config_methods["survivor_selector"](
            self.logger["parent_population_f"],
            self.logger["offspring_population_f"],
            individuals,
        )

    def parent_selection(self, **kwargs) -> np.ndarray:
        """
        ------------------------------------------------------
        Description:
            Apply the parent selection method by the configuration.
        ------------------------------------------------------
        """
        if "parent_selector" not in self.config_methods:
            raise NotImplementedError
        return self.config_methods["parent_selector"](
            self.logger["parent_population_f"]
        )

    # -----------------------------------------------------
    # Private functions.
    # -----------------------------------------------------
    def is_invalid(self, individual: np.ndarray) -> bool:
        """
        ------------------------------------------------------
        Description:
            Check if the current solution is invalid.
        ------------------------------------------------------
        """
        for constraint in self.constraints:
            if not constraint(individual):
                return True
        return False

    def __set_invalid_individuals(self):
        # Fixing solutions and getting aptitude.
        f_offspring = []
        for ind in range(len(self.logger["offspring_population_x"])):
            if self.is_invalid(self.logger["offspring_population_x"][ind]):
                self.logger["offspring_population_x"][ind] = self.fixer(ind)
            f_offspring.append(
                self.aptitude_function(self.logger["offspring_population_x"][ind])
            )
        self.logger["offspring_population_f"] = np.array(f_offspring)

    def __get_pairs(self, parent_ind: np.ndarray):
        parent_ind1 = []
        parent_ind2 = []
        size_ = len(parent_ind) // 2
        ind_individuals_selected = np.random.randint(
            0, len(parent_ind), size=(size_, 2)
        )
        for ind_individual_a, ind_individual_b in ind_individuals_selected:
            parent_ind1.append(parent_ind[ind_individual_a])
            parent_ind2.append(parent_ind[ind_individual_b])

        return parent_ind1, parent_ind2

    def __cross_individuals(
        self, first_parents: np.ndarray, second_parents: np.ndarray
    ) -> None:

        offspring = None
        for parent_a, parent_b in zip(first_parents, second_parents):
            individuals = self.logger["parent_population_x"][[parent_a, parent_b]]

            if np.random.rand() <= self.logger["cross_percentage"]:
                individuals = self.crossover_operator(
                    np.array([parent_a]), np.array([parent_b])
                )
            try:
                offspring = np.concatenate((offspring, individuals), axis=0)
            except ValueError:
                offspring = individuals

        self.logger["offspring_population_x"] = offspring

    def __mutate_individuals(self):
        individuals, variables = self.logger["offspring_population_x"].shape
        individuals_to_mutate = [
            np.random.rand() <= self.logger["mutation_percentage"]
            for i in range(individuals)
        ]
        self.logger["offspring_population_x"][
            individuals_to_mutate
        ] = self.mutation_operator(individuals_to_mutate)



import typing
import logging
from tqdm import tqdm
import numpy as np
import copy
from pyristic.utils.helpers import *

__all__ = ["TabuList", "TabuSearch"]
LOGGER = logging.getLogger()


class TabuList:
    """
    Description:
        Complementary data structure to help the
        tabu search, where save the updates to the current
        solution during a specific time.
    Arguments:
        - timer: integer number (default 2). It indicates
            the time that a item is retained in the tabu list.
    """

    def __init__(self, timer: int = 2):
        self.tabu_list = [[None, None, -1, 100000000]]
        self.timer = timer

    def __str__(self):
        printable_object = "\n ---- Tabu list: ---- \n"
        printable_object += f"List size: {self.tabu_list} \n"
        for position, value, iteration, current_timer in self.tabu_list:
            printable_object += f"p: {position} v:{value} Iteration: {iteration} current timer: {current_timer} \n"
        return printable_object

    def push(self, solution: list) -> None:
        """
        Description:
            Append a solution during a time in the tabu list.
        Arguments:
            - solution:
        """
        assert len(solution) == 3
        solution.append(self.timer)
        self.tabu_list.insert(0, solution)

    def pop_back(self) -> None:
        """
        Description:
            remove a item from the tabu list.
        Arguments:
            - None
        """
        x_last = self.tabu_list.pop()
        x_current_back = self.get_back()
        update_time = self.timer - (x_current_back[2] - x_last[2])
        x_current_back[2] -= update_time
        self.update_back(x_current_back)

    def get_back(self) -> list:
        """
        Description:
            Return the older element in the list.
        Arguments:
            - None.
        """
        return self.tabu_list[-1]

    def update_back(self, solution: list) -> None:
        """
        Description:
            Update the older solution in the list.
        Arguments:
            - None.
        """
        self.tabu_list[-1] = solution

    def find(self, solution: list) -> bool:
        """
        Description:
            Search in the elem
        Arguments:
            - solution: it is an array of two elements. The first position
            is the location in the current solution and the second position is
            the value replaced.
        """
        # X is [p, v], where p is the position changed and v the new value.
        assert len(solution) == 2
        for position, value, *internal_items in self.tabu_list:
            if solution[0] == position and solution[1] == value:
                return True
        return False

    def reset(self, timer: int) -> None:
        """
        Description:
            Clean up the data structure.
        Arguments:
            - None.
        """
        self.tabu_list = []
        self.timer = timer

    def update(self) -> None:
        """
        Description:
            Update the data structure, it controls if the older solution
            should be in the container yet.
        Arguments:
            - None.
        """
        most_old_candidate = copy.deepcopy(self.get_back())
        most_old_candidate[-1] -= 1

        if most_old_candidate[-1] <= 0:
            self.pop_back()
        else:
            self.update_back(most_old_candidate)


class TabuSearch:
    """
    ------------------------------------------------------
    Description:
        Optimization search algorithm inspired on Tabu
        algorithm.
    Arguments:
        - Function: The aptitude function to minimize.
        - Constraints: Array with boolean functions.
        - tabu_struct: Include the structure that helps you save relevant
            informationn during the search (optional argument).
    ------------------------------------------------------
    """

    def __init__(
        self,
        function: typing.Callable[[np.ndarray], typing.Union[int, float]],
        constraints: list,
        tabu_struct=TabuList(),
    ):

        self.tabu_list = tabu_struct  # Initialize tabulist
        self.constraints = constraints
        self.function = function

        # Search information.
        self.logger = {}
        self.logger["best_individual"] = None
        self.logger["best_f"] = None
        self.logger["current_iter"] = None
        self.logger["total_iter"] = None

    def __str__(self):
        printable = (
            "Tabu search: \n "
            f"f(X) = {self.logger['best_f']} \n X = {self.logger['best_individual']} \n "
        )
        first = True

        for constraint in self.constraints:
            if constraint.__doc__ is not None:

                if first:
                    first = False
                    printable += "Constraints: \n "

                constraint(self.logger["best_individual"])
                printable += f"{constraint.__doc__} \n"

        return printable

    def optimize(
        self,
        initial_solution: typing.Union[np.ndarray, typing.Callable[[], np.ndarray]],
        iterations: int,
        memory_time: int,
        verbose: bool = True,
        **kwargs,
    ) -> None:
        """
        ------------------------------------------------------
        Description:
            The main function to find the best solution using tabu search.
        Arguments:
            - Init: Numpy array, represent the initial solution or function which generates a random
              initial solution (this solution should be a numpy array).
            - Iterations:  Integer, stop condition.
            - Memory time: Integer, time which a solution is in the tabu list
        ------------------------------------------------------
        """
        try:
            best_candidate = initial_solution()
        except TypeError:
            best_candidate = copy.deepcopy(initial_solution)

        self.tabu_list.reset(memory_time)

        self.logger["best_individual"] = None
        self.logger["best_f"] = None

        f_candidate = self.function(best_candidate)

        try:
            for step_ in tqdm(range(1, iterations + 1), disable=not verbose):

                neighbors = [
                    neighbor
                    for neighbor in self.get_neighbors(best_candidate, **kwargs)
                    if not self.tabu_list.find(
                        self.encode_change(neighbor, best_candidate, **kwargs)
                    )
                ]
                neighbors = np.array(neighbors)

                valid_neighbors = neighbors[
                    np.apply_along_axis(self.is_valid, 1, neighbors), :
                ]

                # Check if there exists a valid neighbor
                if len(valid_neighbors) == 0:
                    continue

                f_feasible_candidates = np.apply_along_axis(
                    self.function, 1, valid_neighbors
                )

                ind_min = np.argmin(f_feasible_candidates)

                position, value = self.encode_change(
                    valid_neighbors[ind_min], best_candidate, **kwargs
                )
                self.tabu_list.push([position, value, step_])

                if f_feasible_candidates[ind_min] < f_candidate:
                    best_candidate = copy.deepcopy(valid_neighbors[ind_min])
                    f_candidate = f_feasible_candidates[ind_min]

                self.tabu_list.update()

        except KeyboardInterrupt:
            LOGGER.error("Interrupted, saving best solution found so far.")

        self.logger["best_individual"] = best_candidate
        self.logger["best_f"] = f_candidate

    def is_valid(self, solution: np.ndarray) -> bool:
        """
        ------------------------------------------------------
        Description:
            Check if the current solution is valid.
        ------------------------------------------------------
        """
        for constraint in self.constraints:
            if not constraint(solution):
                return False
        return True

    def get_neighbors(self, solution: typing.Union[list, np.ndarray], **kwargs) -> list:
        """
        ------------------------------------------------------
        Description:
            The user has to custom this function.
            The function has to return a list with all possible neighbors.
            Check the examples for a detailed explanation.
        ------------------------------------------------------
        """
        raise NotImplementedError

    def encode_change(
        self,
        neighbor: typing.Union[list, np.ndarray],
        solution: typing.Union[list, np.ndarray],
        **kwargs,
    ) -> list:
        """
        ------------------------------------------------------
        Description:
            The user has to custom this function.
            Check the examples for a detailed explanation.
        ------------------------------------------------------
        """
        raise NotImplementedError



# from distutils.core import setup
from setuptools import find_packages, setup

with open("README.md", encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()


setup(
    name="pyristic",
    version="v1.4.0",
    license="MIT",
    description="Set of metaheuristic for solve optimization problems.",
    author="Jesús Armando Ortíz Peñafiel",
    author_email="armandopenafiel12@gmail.com",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/JAOP1/",
    download_url="https://github.com/JAOP1/pyristic/archive/refs/tags/v1.4.0.tar.gz",
    keywords=["Optimization", "Metaheuristic", "Python"],
    install_requires=[
        "numpy>=1.15.4",
        "tqdm>=4.28.1",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(exclude=("examples", "docs")),
)



import numpy as np

__all__ = [
    "DiscreteCrossover",
    "IntermediateCrossover",
    "NPointCrossover",
    "UniformCrossover",
    "PermutationOrderCrossover",
    "SimulatedBinaryCrossover",
    "discrete_crossover",
    "intermediate_crossover",
    "n_point_crossover",
    "uniform_crossover",
    "permutation_order_crossover",
    "simulated_binary_crossover",
    "NoneCrossover",
]

"""
---------------------------------------------------------------------------------
                                Evolution Strategy.
---------------------------------------------------------------------------------
"""

# Crossover operators.
def discrete_crossover(
    population: np.ndarray, parent_ind1: np.ndarray, parent_ind2: np.ndarray
) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Function which create new elements combining two elements.
        X_{i,s} if v == 1
        X_{i,t} if v == 0
    Arguments:
        -population: Matrix with size m x n, where m is the current size of
        population and n is the problem variables. Every row is an element.
        -parent_ind1: the chosen parents, where every position is the index in X.
        -parent_ind2: the chosen parents, where every position is the index in X.
    ------------------------------------------------------
    """

    rows, cols = population.shape
    rows_new = len(parent_ind1)
    components_selected_fist_parent = np.zeros((rows_new, cols))
    components_selected_second_parent = np.zeros((rows_new, cols))

    for i_row in range(rows_new):
        for j_col in range(cols):
            if np.random.randint(2):
                components_selected_fist_parent[i_row, j_col] = 1
            else:
                components_selected_second_parent[i_row, j_col] = 1

    return (
        population[parent_ind1] * components_selected_fist_parent
        + population[parent_ind2] * components_selected_second_parent
    )


def intermediate_crossover(
    population: np.ndarray,
    parent_ind1: np.ndarray,
    parent_ind2: np.ndarray,
    alpha: float = 0.5,
) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Function which create new elements combining  the average of two elements.
    Arguments:
        -population: Matrix with size m x n, where m is the current size of
        population and n is the problem variables. Every row is an element.
        -parent_ind1: the chosen parents, where every position is the index in X.
        -parent_ind2: the chosen parents, where every position is the index in X.
    ------------------------------------------------------
    """
    return alpha * population[parent_ind1] + (1 - alpha) * population[parent_ind2]


"""
---------------------------------------------------------------------------------
                                Genetic Algorithm.
---------------------------------------------------------------------------------
"""


def n_point_crossover(
    population: np.ndarray,
    parent_ind1: np.ndarray,
    parent_ind2: np.ndarray,
    n_cross: int = 1,
) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Function which create new elements combining two individuals.
    Arguments:
        -population: Matrix with size m x n, where m is the current size of
        population and n is the problem variables. Every row is an element.
        -parent_ind1: the chosen parents, where every position is the index in X.
        -parent_ind2: the chosen parents, where every position is the index in X.
        -n_cross: integer value which says you the number of cuts (by default is 1).

        Note:
            parent_ind1 and parent_ind2 should be m elements.
    Return:
        Matriz with size 2m x n.
    ------------------------------------------------------
    """
    n_cross = int(n_cross)
    num_individuals = len(parent_ind1)
    decision_variables = len(population[0])
    new_population = np.ones(
        (num_individuals * 2, decision_variables), dtype=np.float64
    )

    for i in range(0, num_individuals * 2):
        if i % 2 == 0:
            new_population[i] = population[parent_ind1[(i // 2)]]
        else:
            new_population[i] = population[parent_ind2[(i // 2)]]

    for row in range(num_individuals):
        cross_points = np.random.choice(decision_variables - 1, n_cross, replace=False)
        cross_points += 1
        cross_points.sort()
        for i in range(0, n_cross, 2):
            start = cross_points[i]
            final = decision_variables
            if i + 1 < n_cross:
                final = cross_points[i + 1]
            new_population[2 * row][start:final] = population[parent_ind2[row]][
                start:final
            ]
            new_population[2 * row + 1][start:final] = population[parent_ind1[row]][
                start:final
            ]

    return new_population


def uniform_crossover(
    population: np.ndarray,
    parent_ind1: np.ndarray,
    parent_ind2: np.ndarray,
    flip_prob: int = 0.5,
) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Function which create new elements combining two individuals.
    Arguments:
        -population: Matrix with size m x n, where m is the current size of
        population and n is the problem variables. Every row is an element.
        -parent_ind1: the chosen parents, where every position is the index in X.
        -parent_ind2: the chosen parents, where every position is the index in X.
        -flip_prob: float value which says the likely to change a position (by default is 0.5).
        Note:
            parent_ind1 and parent_ind2 should be m elements.
    Return:
        Matriz with size 2m x n.
    ------------------------------------------------------
    """

    num_individuals = len(parent_ind1)
    decision_variables = len(population[0])
    new_population = np.ones(
        (num_individuals * 2, decision_variables), dtype=np.float64
    )

    for i in range(0, num_individuals * 2):
        if i % 2 == 0:
            new_population[i] = population[parent_ind1[(i // 2)]]
        else:
            new_population[i] = population[parent_ind2[(i // 2)]]

    for i in range(num_individuals):
        for decision_var in range(decision_variables):
            if np.random.rand() <= flip_prob:
                new_population[2 * i][decision_var] = population[parent_ind2[i]][
                    decision_var
                ]
                new_population[2 * i + 1][decision_var] = population[parent_ind1[i]][
                    decision_var
                ]

    return new_population


def permutation_order_crossover(
    population: np.ndarray, parent_ind1: np.ndarray, parent_ind2: np.ndarray
) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Function which create new elements combining two individuals.
    Arguments:
        -population: Matrix with size m x n, where m is the current size of
        population and n is the problem variables. Every row is an element.
        -parent_ind1: the chosen parents, where every position is the index in X.
        -parent_ind2: the chosen parents, where every position is the index in X.
        Note:
            parent_ind1 and parent_ind2 should be m elements.
    Return:
        Matriz with size 2m x n.
    ------------------------------------------------------
    """
    num_individuals = len(parent_ind1)
    decision_variables = len(population[0])
    new_population = np.ones((num_individuals * 2, decision_variables))

    def create_child(parent1: np.ndarray, parent2: np.ndarray) -> np.ndarray:

        interval = np.random.choice(decision_variables + 1, 2, replace=False)
        interval.sort()

        individual = np.full(decision_variables, np.inf)
        segment = parent1[interval[0] : interval[1]]
        individual[interval[0] : interval[1]] = segment
        remainder = []
        for component in parent2:
            if component not in segment:
                remainder.append(component)

        individual[individual == np.inf] = remainder
        return individual

    for i in range(num_individuals):
        new_population[2 * i] = create_child(
            population[parent_ind1[i]], population[parent_ind2[i]]
        )
        new_population[2 * i + 1] = create_child(
            population[parent_ind2[i]], population[parent_ind1[i]]
        )

    return new_population


def simulated_binary_crossover(
    population: np.ndarray,
    parent_ind1: np.ndarray,
    parent_ind2: np.ndarray,
    nc: int = 1,
) -> np.ndarray:
    """
    ------------------------------------------------------
    Description:
        Function which create new elements combining two individuals.
    Arguments:
        -population: Matrix with size m x n, where m is the current size of
        population and n is the problem variables. Every row is an element.
        -parent_ind1: the chosen parents, where every position is the index in X.
        -parent_ind2: the chosen parents, where every position is the index in X.
        -nc: integer number. Default is 1.
        Note:
            parent_ind1 and parent_ind2 should be m elements.
    Return:
        Matriz with size 2m x n.
    ------------------------------------------------------
    """
    nc = int(nc)
    random_float_number = np.random.rand()
    B = -1
    exponent = 1 / (nc + 1)
    if random_float_number <= 0.5:
        B = np.power((2 * random_float_number), exponent)
    else:
        B = np.power(1 / (2 * (1 - random_float_number)), exponent)

    first_parent_population = population[parent_ind1]
    second_parent_population = population[parent_ind2]

    A = first_parent_population + second_parent_population
    B = np.absolute(second_parent_population - first_parent_population) * B

    return np.concatenate((0.5 * (A - B), 0.5 * (A + B)), axis=0)


class IntermediateCrossover:
    """
    Description:
      Class crossover operator based on intermediate_crossover.
    Arguments:
        - alpha: Float number (default = 0.5).
    """

    def __init__(self, alpha: float = 0.5):
        self.alpha = alpha
        self.__doc__ = f"Intermediate\n\tArguments:\n\t\t-Alpha:{self.alpha}"

    def __call__(
        self, population: np.ndarray, parent_ind1: np.ndarray, parent_ind2: np.ndarray
    ) -> np.ndarray:
        assert len(parent_ind1) == len(parent_ind2)
        return intermediate_crossover(population, parent_ind1, parent_ind2, self.alpha)


class NPointCrossover:
    """
    Description:
      Class crossover operator based on the 1 point crossover. This is a
      generic crossover operator for n partitions.
    Arguments:
        - n: Integer number (default = 1).
    """

    def __init__(self, n_cross: int = 1):
        self.n_cross = n_cross
        self.__doc__ = f"n point\n\tArguments:\n\t\t-n_cross: {self.n_cross}"

    def __call__(
        self, population: np.ndarray, parent_ind1: np.ndarray, parent_ind2: np.ndarray
    ) -> np.ndarray:
        assert len(parent_ind1) == len(parent_ind2)

        return n_point_crossover(population, parent_ind1, parent_ind2, self.n_cross)


class UniformCrossover:
    """
    Description:
      Class crossover operator based on the uniform crossover.
    Arguments:
        - flip_prob: float number between [0,1] (default = 0.5).
    """

    def __init__(self, flip_prob: float = 0.5):
        self.prob = flip_prob
        self.__doc__ = f"Uniform\n\tArguments:\n\t\t-prob: {self.prob}"

    def __call__(
        self, population: np.ndarray, parent_ind1: np.ndarray, parent_ind2: np.ndarray
    ) -> np.ndarray:
        assert len(parent_ind1) == len(parent_ind2)
        return uniform_crossover(population, parent_ind1, parent_ind2, self.prob)


class PermutationOrderCrossover:
    """
    Description:
      Class crossover operator based on the permutation order crossover.
    Arguments:
        - None
    """

    def __init__(self):
        self.__doc__ = "Permutation order"

    def __call__(
        self, population: np.ndarray, parent_ind1: np.ndarray, parent_ind2: np.ndarray
    ) -> np.ndarray:
        assert len(parent_ind1) == len(parent_ind2)
        return permutation_order_crossover(population, parent_ind1, parent_ind2)


class SimulatedBinaryCrossover:
    """
    Description:
      Class crossover operator based on the simulated binary crossover.
    Arguments:
        - nc: Integer number (default = 1).
    """

    def __init__(self, nc: int = 1):
        self.nc = nc
        self.__doc__ = f"Simulated binary\n\tArguments:\n\t\t-nc: {self.nc}"

    def __call__(
        self, population: np.ndarray, parent_ind1: np.ndarray, parent_ind2: np.ndarray
    ) -> np.ndarray:
        assert len(parent_ind1) == len(parent_ind2)

        return simulated_binary_crossover(population, parent_ind1, parent_ind2, self.nc)


class NoneCrossover:
    """
    Description:
      Class crossover operator to avoid include crossover operator.
    Arguments:
        - None
    """

    def __init__(self):
        self.__doc__ = "None"

    def __call__(self, population: np.ndarray) -> np.ndarray:
        return population


# Crossover operator for ES.
class DiscreteCrossover:
    """
    Description:
      Class crossover operator based on the discrete crossover.
    Arguments:
        - None
    """

    def __init__(self):
        self.__doc__ = "Discrete"

    def __call__(
        self, population: np.ndarray, parent_ind1: np.ndarray, parent_ind2: np.ndarray
    ) -> np.ndarray:
        assert len(parent_ind1) == len(parent_ind2)
        return discrete_crossover(population, parent_ind1, parent_ind2)
